@@ -20,8 +20,10 @@ import os
 import sys
 import uuid
 
-from bson import json_util, BSON
-from bson.binary import Binary, STANDARD
+import bson
+from bson.raw_bson import RawBSONDocument
+from bson import json_util
+from bson.binary import Binary, UuidRepresentation
 from bson.codec_options import CodecOptions
 from bson.json_util import JSONOptions
 from bson.son import SON
@@ -30,7 +32,7 @@ sys.path[0:0] = [""]
 
 from pymongocrypt.auto_encrypter import AutoEncrypter
 from pymongocrypt.binding import lib
-from pymongocrypt.compat import unicode_type, PY3
+from pymongocrypt.compat import unicode_type, safe_bytearray_or_base64, PY3
 from pymongocrypt.errors import MongoCryptError
 from pymongocrypt.explicit_encrypter import ExplicitEncrypter
 from pymongocrypt.mongocrypt import (MongoCrypt,
@@ -65,6 +67,12 @@ class TestMongoCryptBinary(unittest.TestCase):
             self.assertEqual(binary.to_bytes(), b'')
         self.assertIsNone(binary.bin)
 
+        # Memoryview
+        with MongoCryptBinaryIn(memoryview(b'1\x0023')) as binary:
+            self.assertIsNotNone(binary.bin)
+            self.assertEqual(binary.to_bytes(), b'1\x0023')
+        self.assertIsNone(binary.bin)
+
     def test_mongocrypt_binary_out(self):
         with MongoCryptBinaryOut() as binary:
             self.assertIsNotNone(binary.bin)
@@ -97,13 +105,20 @@ class TestMongoCryptOptions(unittest.TestCase):
             opts = MongoCryptOptions(kms_providers, schema_map)
             self.assertEqual(opts.kms_providers, kms_providers, msg=kms_providers)
             self.assertEqual(opts.schema_map, schema_map)
+            self.assertIsNone(opts.encrypted_fields_map)
+            self.assertFalse(opts.bypass_query_analysis)
+
+        encrypted_fields_map = bson_data('encrypted-field-config-map.json')
+        opts = MongoCryptOptions(valid[0][0], schema_map, encrypted_fields_map=encrypted_fields_map,
+                                 bypass_query_analysis=True)
+        self.assertEqual(opts.encrypted_fields_map, encrypted_fields_map)
+        self.assertTrue(opts.bypass_query_analysis)
 
     def test_mongocrypt_options_validation(self):
         with self.assertRaisesRegex(
                 ValueError, 'at least one KMS provider must be configured'):
             MongoCryptOptions({})
         for invalid_kms_providers in [
-                {'aws': {}},
                 {'aws': {'accessKeyId': 'foo'}},
                 {'aws': {'secretAccessKey': 'foo'}}]:
             with self.assertRaisesRegex(
@@ -125,8 +140,13 @@ class TestMongoCryptOptions(unittest.TestCase):
                 TypeError, "schema_map must be bytes or None"):
             MongoCryptOptions(valid_kms, schema_map={})
 
+        with self.assertRaisesRegex(
+                TypeError, "encrypted_fields_map must be bytes or None"):
+            MongoCryptOptions(valid_kms, encrypted_fields_map={})
+
 
 class TestMongoCrypt(unittest.TestCase):
+    maxDiff = None
 
     def test_mongocrypt(self):
         kms_providers = {
@@ -208,10 +228,10 @@ class TestMongoCrypt(unittest.TestCase):
             MongoCrypt(options, callback)
 
     @staticmethod
-    def create_mongocrypt():
+    def create_mongocrypt(**kwargs):
         return MongoCrypt(MongoCryptOptions({
             'aws': {'accessKeyId': 'example', 'secretAccessKey': 'example'},
-            'local': {'key': b'\x00'*96}}), MockCallback())
+            'local': {'key': b'\x00'*96}}, **kwargs), MockCallback())
 
     def _test_kms_context(self, ctx):
         key_filter = ctx.mongo_operation()
@@ -223,18 +243,22 @@ class TestMongoCrypt(unittest.TestCase):
         km_contexts = list(ctx.kms_contexts())
         self.assertEqual(len(km_contexts), 1)
         with km_contexts[0] as kms_ctx:
-            self.assertEqual(kms_ctx.endpoint, "kms.us-east-1.amazonaws.com")
+            self.assertEqual(kms_ctx.kms_provider, "aws")
+            self.assertEqual(kms_ctx.endpoint, "kms.us-east-1.amazonaws.com:443")
             self.assertEqual(len(kms_ctx.message), 781)
             self.assertEqual(kms_ctx.bytes_needed, 1024)
 
             kms_ctx.feed(http_data('kms-reply.txt'))
             self.assertEqual(kms_ctx.bytes_needed, 0)
+            self.assertEqual(kms_ctx.kms_provider, "aws")
 
         ctx.complete_kms()
 
     def test_encrypt(self):
         mc = self.create_mongocrypt()
         self.addCleanup(mc.close)
+        if mc.crypt_shared_lib_version is not None:
+            self.skipTest("this test must be skipped when crypt_shared is loaded")
         with mc.encryption_context('text', bson_data('command.json')) as ctx:
             self.assertEqual(ctx.state, lib.MONGOCRYPT_CTX_NEED_MONGO_COLLINFO)
 
@@ -247,6 +271,8 @@ class TestMongoCrypt(unittest.TestCase):
             self.assertEqual(ctx.state, lib.MONGOCRYPT_CTX_NEED_MONGO_MARKINGS)
 
             mongocryptd_cmd = ctx.mongo_operation()
+            self.assertEqual(bson.decode(mongocryptd_cmd, OPTS),
+                             json_data('mongocryptd-command.json'))
             self.assertEqual(mongocryptd_cmd,
                              bson_data('mongocryptd-command.json'))
 
@@ -259,8 +285,8 @@ class TestMongoCrypt(unittest.TestCase):
             self.assertEqual(ctx.state, lib.MONGOCRYPT_CTX_READY)
 
             encrypted = ctx.finish()
-            self.assertEqual(
-                BSON(encrypted).decode(), json_data('encrypted-command.json'))
+            self.assertEqual(bson.decode(encrypted, OPTS),
+                             json_data('encrypted-command.json'))
             self.assertEqual(encrypted, bson_data('encrypted-command.json'))
             self.assertEqual(ctx.state, lib.MONGOCRYPT_CTX_DONE)
 
@@ -276,11 +302,39 @@ class TestMongoCrypt(unittest.TestCase):
             self.assertEqual(ctx.state, lib.MONGOCRYPT_CTX_READY)
 
             encrypted = ctx.finish()
-            self.assertEqual(
-                BSON(encrypted).decode(), json_data('command-reply.json'))
+            self.assertEqual(bson.decode(encrypted, OPTS),
+                             json_data('command-reply.json'))
             self.assertEqual(encrypted, bson_data('command-reply.json'))
             self.assertEqual(ctx.state, lib.MONGOCRYPT_CTX_DONE)
 
+    def test_encrypt_encrypted_fields_map(self):
+        encrypted_fields_map = bson_data('compact/success/encrypted-field-config-map.json')
+        mc = self.create_mongocrypt(encrypted_fields_map=encrypted_fields_map)
+        self.addCleanup(mc.close)
+        with mc.encryption_context('db', bson_data('compact/success/cmd.json')) as ctx:
+            self.assertEqual(ctx.state, lib.MONGOCRYPT_CTX_NEED_MONGO_KEYS)
+
+            ctx.mongo_operation()
+            ctx.add_mongo_operation_result(bson_data(
+                'keys/12345678123498761234123456789012-local-document.json'))
+            self.assertEqual(ctx.state, lib.MONGOCRYPT_CTX_NEED_MONGO_KEYS)
+            ctx.mongo_operation()
+            ctx.add_mongo_operation_result(bson_data(
+                'keys/ABCDEFAB123498761234123456789012-local-document.json'))
+            self.assertEqual(ctx.state, lib.MONGOCRYPT_CTX_NEED_MONGO_KEYS)
+            ctx.mongo_operation()
+            ctx.add_mongo_operation_result(bson_data(
+                'keys/12345678123498761234123456789013-local-document.json'))
+            ctx.complete_mongo_operation()
+
+            self.assertEqual(ctx.state, lib.MONGOCRYPT_CTX_READY)
+
+            encrypted = ctx.finish()
+            self.assertEqual(bson.decode(encrypted, OPTS),
+                             json_data('compact/success/encrypted-payload.json'))
+            self.assertEqual(encrypted, bson_data('compact/success/encrypted-payload.json'))
+            self.assertEqual(ctx.state, lib.MONGOCRYPT_CTX_DONE)
+
 
 class MockCallback(MongoCryptCallback):
     def __init__(self,
@@ -311,13 +365,14 @@ class MockCallback(MongoCryptCallback):
         raise NotImplementedError
 
     def bson_encode(self, doc):
-        return BSON.encode(doc)
+        return bson.encode(doc)
 
     def close(self):
         pass
 
 
 class TestMongoCryptCallback(unittest.TestCase):
+    maxDiff = None
 
     @staticmethod
     def mongo_crypt_opts():
@@ -325,6 +380,31 @@ class TestMongoCryptCallback(unittest.TestCase):
             'aws': {'accessKeyId': 'example', 'secretAccessKey': 'example'},
             'local': {'key': b'\x00'*96}})
 
+    @unittest.skipUnless(os.getenv("TEST_CRYPT_SHARED"), "this test requires TEST_CRYPT_SHARED=1")
+    def test_crypt_shared(self):
+        kms_providers = {
+            'aws': {'accessKeyId': 'example', 'secretAccessKey': 'example'},
+            'local': {'key': b'\x00'*96}}
+        mc = MongoCrypt(MongoCryptOptions(kms_providers), MockCallback())
+        self.addCleanup(mc.close)
+        self.assertIsNotNone(mc.crypt_shared_lib_version)
+        # Test that we can pick up crypt_shared automatically
+        encrypter = AutoEncrypter(MockCallback(), MongoCryptOptions(
+            kms_providers,
+            bypass_encryption=False,
+            crypt_shared_lib_required=True))
+        self.addCleanup(encrypter.close)
+        encrypter = AutoEncrypter(MockCallback(), MongoCryptOptions(
+            kms_providers,
+            crypt_shared_lib_path=os.environ["CRYPT_SHARED_PATH"],
+            crypt_shared_lib_required=True))
+        self.addCleanup(encrypter.close)
+        with self.assertRaisesRegex(MongoCryptError, "/doesnotexist"):
+            AutoEncrypter(MockCallback(), MongoCryptOptions(
+                kms_providers,
+                crypt_shared_lib_path="/doesnotexist",
+                crypt_shared_lib_required=True))
+
     def test_encrypt(self):
         encrypter = AutoEncrypter(MockCallback(
             list_colls_result=bson_data('collection-info.json'),
@@ -332,9 +412,9 @@ class TestMongoCryptCallback(unittest.TestCase):
             key_docs=[bson_data('key-document.json')],
             kms_reply=http_data('kms-reply.txt')), self.mongo_crypt_opts())
         self.addCleanup(encrypter.close)
-        encrypted = encrypter.encrypt('text', bson_data('command.json'))
-        self.assertEqual(
-            BSON(encrypted).decode(), json_data('encrypted-command.json'))
+        encrypted = encrypter.encrypt('test', bson_data('command.json'))
+        self.assertEqual(bson.decode(encrypted, OPTS),
+                         json_data('encrypted-command.json'))
         self.assertEqual(encrypted, bson_data('encrypted-command.json'))
 
     def test_decrypt(self):
@@ -346,8 +426,8 @@ class TestMongoCryptCallback(unittest.TestCase):
         self.addCleanup(encrypter.close)
         decrypted = encrypter.decrypt(
             bson_data('encrypted-command-reply.json'))
-        self.assertEqual(
-            BSON(decrypted).decode(), json_data('command-reply.json'))
+        self.assertEqual(bson.decode(decrypted, OPTS),
+                         json_data('command-reply.json'))
         self.assertEqual(decrypted, bson_data('command-reply.json'))
 
 
@@ -361,10 +441,11 @@ class KeyVaultCallback(MockCallback):
 
     def insert_data_key(self, data_key):
         self.data_key = data_key
-        return BSON(data_key).decode()['_id']
+        return bson.decode(data_key, OPTS)['_id']
 
 
 class TestExplicitEncryption(unittest.TestCase):
+    maxDiff = None
 
     @staticmethod
     def mongo_crypt_opts():
@@ -379,26 +460,82 @@ class TestExplicitEncryption(unittest.TestCase):
         self.addCleanup(encrypter.close)
 
         val = {'v': 'hello'}
-        encoded_val = BSON.encode(val)
+        encoded_val = bson.encode(val)
         algo = "AEAD_AES_256_CBC_HMAC_SHA_512-Deterministic"
         encrypted = encrypter.encrypt(
             encoded_val, algo, key_id=key_id, key_alt_name=key_alt_name)
-        self.assertEqual(
-            BSON(encrypted).decode(), json_data('encrypted-value.json'))
+        self.assertEqual(bson.decode(encrypted, OPTS),
+                         json_data('encrypted-value.json'))
         self.assertEqual(encrypted, bson_data('encrypted-value.json'))
 
         decrypted = encrypter.decrypt(encrypted)
-        self.assertEqual(BSON(decrypted).decode(), val)
+        self.assertEqual(bson.decode(decrypted, OPTS), val)
         self.assertEqual(encoded_val, decrypted)
 
     def test_encrypt_decrypt(self):
         key_id = json_data('key-document.json')['_id']
-        self._test_encrypt_decrypt(key_id=key_id.bytes)
+        self._test_encrypt_decrypt(key_id=key_id)
 
     def test_encrypt_decrypt_key_alt_name(self):
         key_alt_name = json_data('key-document.json')['keyAltNames'][0]
         self._test_encrypt_decrypt(key_alt_name=key_alt_name)
 
+    def test_encrypt_errors(self):
+        key_id = json_data('key-document.json')['_id']
+        encrypter = ExplicitEncrypter(MockCallback(key_docs=[]), self.mongo_crypt_opts())
+        self.addCleanup(encrypter.close)
+
+        val = {'v': 'value123'}
+        encoded_val = bson.encode(val)
+        # Invalid algorithm.
+        with self.assertRaisesRegex(MongoCryptError, "algorithm"):
+            encrypter.encrypt(encoded_val, "Invalid", key_id)
+        # Invalid query_type type.
+        with self.assertRaisesRegex(TypeError, "query_type"):
+            encrypter.encrypt(encoded_val, "Indexed", key_id, query_type=42)
+        # Invalid query_type string.
+        with self.assertRaisesRegex(MongoCryptError, "query_type"):
+            encrypter.encrypt(encoded_val, "Indexed", key_id, query_type='invalid query type string')
+        # Invalid contention_factor type.
+        with self.assertRaisesRegex(TypeError, "contention_factor"):
+            encrypter.encrypt(encoded_val, "Indexed", key_id, contention_factor='not an int')
+        with self.assertRaisesRegex(MongoCryptError, "contention"):
+            encrypter.encrypt(encoded_val, "Indexed", key_id, contention_factor=-1)
+        # Invalid: Unindexed + query_type is an error.
+        with self.assertRaisesRegex(MongoCryptError, "query"):
+            encrypter.encrypt(encoded_val, "Unindexed", key_id, query_type='equality')
+        # Invalid: Unindexed + contention_factor is an error.
+        with self.assertRaisesRegex(MongoCryptError, "contention"):
+            encrypter.encrypt(encoded_val, "Unindexed", key_id, contention_factor=1)
+
+    def test_encrypt_indexed(self):
+        key_path = 'keys/ABCDEFAB123498761234123456789012-local-document.json'
+        key_id = json_data(key_path)['_id']
+        encrypter = ExplicitEncrypter(MockCallback(
+            key_docs=[bson_data(key_path)],
+            kms_reply=http_data('kms-reply.txt')), self.mongo_crypt_opts())
+        self.addCleanup(encrypter.close)
+
+        val = {'v': 'value123'}
+        encoded_val = bson.encode(val)
+        for kwargs in [
+            dict(algorithm='Indexed', contention_factor=0),
+            dict(algorithm='Indexed', query_type='equality', contention_factor=0),
+            dict(algorithm='Indexed', contention_factor=100),
+            dict(algorithm='Unindexed'),
+        ]:
+            kwargs['key_id'] = key_id
+            encrypted = encrypter.encrypt(encoded_val, **kwargs)
+            encrypted_val = bson.decode(encrypted, OPTS)['v']
+            self.assertIsInstance(encrypted_val, Binary)
+            self.assertEqual(encrypted_val.subtype, 6)
+
+            # Queryable Encryption find payloads cannot be round-tripped.
+            if 'query_type' not in kwargs:
+                decrypted = encrypter.decrypt(encrypted)
+                self.assertEqual(bson.decode(decrypted, OPTS), val)
+                self.assertEqual(encoded_val, decrypted)
+
     def test_data_key_creation(self):
         mock_key_vault = KeyVaultCallback(
             kms_reply=http_data('kms-encrypt-reply.txt'))
@@ -418,8 +555,9 @@ class TestExplicitEncryption(unittest.TestCase):
             key_id = encrypter.create_data_key(
                 kms_provider, master_key=master_key,
                 key_alt_names=key_alt_names)
-            self.assertIsInstance(key_id, uuid.UUID)
-            data_key = BSON(mock_key_vault.data_key).decode()
+            self.assertIsInstance(key_id, Binary)
+            self.assertEqual(key_id.subtype, 4)
+            data_key = bson.decode(mock_key_vault.data_key, OPTS)
             # CDRIVER-3277 The order of key_alt_names is not maintained.
             for name in key_alt_names:
                 self.assertIn(name, data_key['keyAltNames'])
@@ -430,19 +568,47 @@ class TestExplicitEncryption(unittest.TestCase):
             "key": "key",
             "endpoint": "example.com"
         }
-        encrypter.create_data_key("aws", master_key=master_key)
-        self.assertEqual("example.com", mock_key_vault.kms_endpoint)
+        key_material = base64.b64decode('xPTAjBRG5JiPm+d3fj6XLi2q5DMXUS/f1f+SMAlhhwkhDRL0kr8r9GDLIGTAGlvC+HVjSIgdL+RKwZCvpXSyxTICWSXTUYsWYPyu3IoHbuBZdmw2faM3WhcRIgbMReU5')
+        if not PY3:
+            key_material = Binary(key_material)
+        encrypter.create_data_key("aws", master_key=master_key, key_material=key_material)
+        self.assertEqual("example.com:443", mock_key_vault.kms_endpoint)
 
+    def test_data_key_creation_bad_key_material(self):
+        mock_key_vault = KeyVaultCallback(
+            kms_reply=http_data('kms-encrypt-reply.txt'))
+        encrypter = ExplicitEncrypter(mock_key_vault, self.mongo_crypt_opts())
+        self.addCleanup(encrypter.close)
+
+        key_material = Binary(b'0' * 97)
+        with self.assertRaisesRegex(MongoCryptError, "keyMaterial should have length 96, but has length 97"):
+            encrypter.create_data_key("local", key_material=key_material)
+
+    def test_rewrap_many_data_key(self):
+        key_path = 'keys/ABCDEFAB123498761234123456789012-local-document.json'
+        key_path2 = 'keys/12345678123498761234123456789012-local-document.json'
+        encrypter = ExplicitEncrypter(MockCallback(
+            key_docs=[bson_data(key_path), bson_data(key_path2)]), self.mongo_crypt_opts())
+        self.addCleanup(encrypter.close)
+
+        result = encrypter.rewrap_many_data_key({})
+        raw_doc = RawBSONDocument(result)
+        assert len(raw_doc['v']) == 2
 
 def read(filename, **kwargs):
     with open(os.path.join(DATA_DIR, filename), **kwargs) as fp:
         return fp.read()
 
 
-OPTS = CodecOptions(uuid_representation=STANDARD)
+OPTS = CodecOptions(uuid_representation=UuidRepresentation.UNSPECIFIED)
 
 # Use SON to preserve the order of fields while parsing json.
-JSON_OPTS = JSONOptions(document_class=SON, uuid_representation=STANDARD)
+if sys.version_info[:2] < (3, 6):
+    document_class = SON
+else:
+    document_class = dict
+JSON_OPTS = JSONOptions(document_class=document_class,
+                        uuid_representation=UuidRepresentation.UNSPECIFIED)
 
 
 def json_data(filename):
@@ -450,7 +616,7 @@ def json_data(filename):
 
 
 def bson_data(filename):
-    return BSON.encode(json_data(filename), codec_options=OPTS)
+    return bson.encode(json_data(filename), codec_options=OPTS)
 
 
 def http_data(filename):
